// Copyright Jiang Du
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <assert.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>
#include <fstream>
#include <boost/lexical_cast.hpp>
#include <boost/random.hpp>
#include <boost/shared_ptr.hpp>
#include "log.h"
#include "reseq_sim.h"
using namespace std;
using namespace boost;
using namespace jsc::util;

int syntax_err() {
	L_(error) << "Usage:\nreseq_sim\n  unit_cost overlap_threshold\n  insertion_info_path map2ref_path_prefix map2target_path_prefix\n  (single_read_type read_length cost_per_bp total_cost)+\n  (pe_read_type pe_read_length pe_gap_mean pe_gap_sd cost_per_bp total_cost)+";
	return 1;
}

int lexical_cast_err() {
	L_(error) << "Lexical_cast error when converting arguments to numeric values";
	return 1;
}

int main(int argc, char * argv[])
{
	// initialize the logging system
	// Log::ReportingLevel = LogLevel::info;
	Log::ReportingLevel = LogLevel::debug2;

	// get parameters
	if (argc < 8) {
		return syntax_err();
	}
	int argi = 1;
	double unit_cost;
	long overlap_threshold;
	string insertion_info_path, map2ref_path_prefix, map2target_path_prefix;
	// for single reads
	vector<string> single_read_type;
	vector<long> read_length;
	vector<double> cost_per_bp, total_cost;
	// for paired-end reads
	vector<string> pe_read_type;
	vector<long> pe_read_length;
	vector<long> pe_gap_mean;
	vector<long> pe_gap_sd;
	vector<double> pe_cost_per_bp, pe_total_cost;

	try {
		unit_cost = lexical_cast<double>(argv[argi++]);
		overlap_threshold = lexical_cast<long>(argv[argi++]);
		insertion_info_path = argv[argi++];
		map2ref_path_prefix = argv[argi++];
		map2target_path_prefix = argv[argi++];

		while (argi < argc) {
			string read_type = argv[argi];
			// Here we assume that all the paired-end read types
			// end with "_PE"
			if (read_type.substr(read_type.size() - 3) != "_PE") {
				if (argc - argi < 4) {
					return syntax_err();
				}
				single_read_type.push_back(argv[argi++]);
				read_length.push_back(lexical_cast<long>(argv[argi++]));
				cost_per_bp.push_back(lexical_cast<double>(argv[argi++]));
				double cost = lexical_cast<double>(argv[argi++]);
				total_cost.push_back(unit_cost * cost);
			} else {
				if (argc - argi < 6) {
					return syntax_err();
				}
				pe_read_type.push_back(argv[argi++]);
				pe_read_length.push_back(lexical_cast<long>(argv[argi++]));
				pe_gap_mean.push_back(lexical_cast<long>(argv[argi++]));
				pe_gap_sd.push_back(lexical_cast<long>(argv[argi++]));
				pe_cost_per_bp.push_back(lexical_cast<double>(argv[argi++]));
				double cost = lexical_cast<double>(argv[argi++]);
				pe_total_cost.push_back(unit_cost * cost);
			}
		}
	}
	catch (bad_lexical_cast &)
	{
		return lexical_cast_err();
	}

	// global variables
	ifstream ifs;
	long insertion_start, insertion_end;
	vmers map_mers_ref, map_mers_target;
	vmers map_mers_snp_ref, map_mers_snp_target;
	vvcounts map_counts_ref, map_counts_target;
	vvcounts map_counts_snp_ref, map_counts_snp_target;
	long region_size;

	// load insertion info
	ifs.clear();
	ifs.open(insertion_info_path.c_str());
	assert(ifs.is_open() && !ifs.eof());
	load_insertion_info(ifs, insertion_start, insertion_end);
	ifs.close();
	L_(info) << "Loaded Insertion: [" << insertion_start << ", "
		<< insertion_end << ")";
	region_size = insertion_end;

	// load map counts to ref
	ifs.clear();
	ifs.open((map2ref_path_prefix + ".mapcount").c_str());
	assert(ifs.is_open() && !ifs.eof());
	load_map_counts(ifs, map_mers_ref, map_counts_ref);
	ifs.close();
	ifs.clear();
	ifs.open((map2ref_path_prefix + ".w_snp.mapcount").c_str());
	assert(ifs.is_open() && !ifs.eof());
	load_map_counts(ifs, map_mers_snp_ref, map_counts_snp_ref);
	ifs.close();
	// load map counts to target
	ifs.clear();
	ifs.open((map2target_path_prefix + ".mapcount").c_str());
	assert(ifs.is_open() && !ifs.eof());
	load_map_counts(ifs, map_mers_target, map_counts_target);
	ifs.close();
	ifs.clear();
	ifs.open((map2target_path_prefix + ".w_snp.mapcount").c_str());
	assert(ifs.is_open() && !ifs.eof());
	load_map_counts(ifs, map_mers_snp_target, map_counts_snp_target);
	ifs.close();

	// prepare the bernoulli rngs
	base_generator_type rng;
	rng.seed(static_cast<unsigned int>(std::time(0)));
	srand(time(0));
	uniform_int<long> unif(0, region_size - 1);
	unifrng pos_rng(rng, unif);

	vector<contig_sets> pe_reads_vec;
	// generate the paired end reads (recording the ``effective" ends only)
	for (unsigned long i = 0; i < pe_read_type.size(); ++i) {
		contig_sets pe_reads;
		uniform_int<long> read_length_unif(0, pe_read_length[i]);
		unifrng end_size_rng(rng, read_length_unif);
		normal_distribution<double> norm_dist(pe_gap_mean[i], pe_gap_sd[i]);
		normalrng gap_size_rng(rng, norm_dist);
		if (pe_read_type[i] == contig_type::TYPES[contig_type::LONG_READ_PE]) {
			generate_pe_reads(contig_type::LONG_READ_PE, pe_reads,
					region_size, pos_rng, end_size_rng, gap_size_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					map_mers_ref, map_counts_ref,
					map_mers_snp_ref, map_counts_snp_ref,
					pe_total_cost[i], pe_cost_per_bp[i], pe_read_length[i],
					pe_gap_mean[i], pe_gap_sd[i]);
		} else if (pe_read_type[i] ==
				contig_type::TYPES[contig_type::MEDIUM_READ_PE]) {
			generate_pe_reads(contig_type::MEDIUM_READ_PE, pe_reads,
					region_size, pos_rng, end_size_rng, gap_size_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					map_mers_ref, map_counts_ref,
					map_mers_snp_ref, map_counts_snp_ref,
					pe_total_cost[i], pe_cost_per_bp[i], pe_read_length[i],
					pe_gap_mean[i], pe_gap_sd[i]);
		} else if (pe_read_type[i] ==
				contig_type::TYPES[contig_type::SHORT_READ_PE]) {
			generate_pe_reads(contig_type::SHORT_READ_PE, pe_reads,
					region_size, pos_rng, end_size_rng, gap_size_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					map_mers_ref, map_counts_ref,
					map_mers_snp_ref, map_counts_snp_ref,
					pe_total_cost[i], pe_cost_per_bp[i], pe_read_length[i],
					pe_gap_mean[i], pe_gap_sd[i]);
		} else {
			L_(error) << "Undefined read type: " << pe_read_type[i];
		}
		print_contig_sets(pe_reads);
		pe_reads_vec.push_back(pe_reads);
	}

	// generate the single reads
	vector<contig_sets> single_reads_vec;
	for (unsigned long i = 0; i < single_read_type.size(); ++i) {
		contig_sets single_reads;
		if (single_read_type[i] == contig_type::TYPES[contig_type::LONG_READ]) {
			generate_reads(contig_type::LONG_READ, single_reads,
					region_size, pos_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					total_cost[i], cost_per_bp[i], read_length[i]);
		} else if (single_read_type[i] ==
				contig_type::TYPES[contig_type::MEDIUM_READ]) {
			contig_sets medium_reads;
			generate_reads(contig_type::MEDIUM_READ, single_reads,
					region_size, pos_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					total_cost[i], cost_per_bp[i], read_length[i]);
		} else if (single_read_type[i] ==
				contig_type::TYPES[contig_type::SHORT_READ]) {
			contig_sets short_reads;
			generate_reads(contig_type::SHORT_READ, single_reads,
					region_size, pos_rng,
					map_mers_target, map_counts_target,
					map_mers_snp_target, map_counts_snp_target,
					total_cost[i], cost_per_bp[i], read_length[i]);
		} else {
			L_(error) << "Undefined read type: " << single_read_type[i];
		}
		print_contig_sets(single_reads);
		single_reads_vec.push_back(single_reads);
	}

	// combine the reads
	contig_sets reads;
	BOOST_FOREACH(contig_sets const & pe_reads, pe_reads_vec) {
		combine_reads(reads, pe_reads);
	}
	BOOST_FOREACH(contig_sets const & single_reads, single_reads_vec) {
		combine_reads(reads, single_reads);
	}
	print_contig_sets(reads);

	// build contigs
	contig_sets contigs;
	build_contigs(contigs, reads, overlap_threshold,
			map_mers_target, map_counts_target,
			pos_rng);
	print_contig_sets(contigs);
	
	// identify insertion
	bool detected = identify_insertion(
			contigs,
			insertion_start, insertion_end,
			map_mers_ref, map_counts_ref);
	L_(info) << "Insertion" << (detected ? "" : " not") << " detected";

	// count detection
	long count_detected = 0;
	count_detection(contigs, insertion_start, insertion_end,
			count_detected);
	count_detected = detected ? count_detected : 0;
	L_(info) << "Detected: " << count_detected << "bp / "
		<< (insertion_end - insertion_start) << "bp";

	// print the reconstruction rate
	cout << (double)count_detected / (double)(insertion_end - insertion_start) << endl;

	return 0;
}
