// Copyright Jiang Du
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef _reseq_sim_h_included_
#define _reseq_sim_h_included_

#include <assert.h>
#include <stdlib.h>
#include <algorithm>
#include <cctype>
#include <fstream>
#include <functional>
#include <iostream>
#include <map>
#include <numeric>
#include <set>
#include <sstream>
#include <vector>
#include <sstream>
#include <string>
#include <boost/config.hpp>
#include <boost/foreach.hpp>
#include <boost/random.hpp>
#include <boost/shared_ptr.hpp>
#include "log.h"
using namespace std;
using namespace boost;
using namespace jsc::util;

class constants {
	public:
		static const long min_possible_unique_kmer_size = 20;
		static const long min_pe_based_contig_max_pe_length = 30;
};

class contig_type {
	public:
		static const int LONG_READ = 0;
		static const int MEDIUM_READ = 1;
		static const int SHORT_READ = 2;
		static const int SHORT_CONTIG = 3;
		static const int CONTIG = 4;
		static const int LONG_READ_PE = 5;
		static const int MEDIUM_READ_PE = 6;
		static const int SHORT_READ_PE = 7;
		static const string TYPES[];
};
// Note that all the paired-end read types end with "_PE"
// for the convenience of distinguishing them from single read types
const string contig_type::TYPES[] = {"LONG_READ", "MEDIUM_READ", "SHORT_READ", "SHORT_CONTIG", "CONTIG", "LONG_READ_PE", "MEDIUM_READ_PE", "SHORT_READ_PE"};

class contig_set {	// a group of contigs with the same type
	public:
		int type;
		long start;
		long length;
		long max_pe_size;
		vector<long> sames;
		vector<long> similars;
		vector<long> substitutions;
		vector<long> insertions;
		vector<long> deletions;
		contig_set(int const type, long const & start,
				long const & length) {
			this->type = type;
			this->start = start;
			this->length = length;
			this->max_pe_size = 0;
			this->sames.assign(length, 0);
			this->similars.assign(length, 0);
			this->substitutions.assign(length, 0);
			this->insertions.assign(length, 0);
			this->deletions.assign(length, 0);
		}

		void resize(long const & length) {
			this->length = length;
			this->sames.resize(length, 0);
			this->similars.resize(length, 0);
			this->substitutions.resize(length, 0);
			this->insertions.resize(length, 0);
			this->deletions.resize(length, 0);
		}
};

typedef vector<shared_ptr<contig_set> > contig_sets;
typedef vector<shared_ptr<contig_sets> > regional_contig_sets;

typedef vector<long> vmers;
typedef vector<long> vcounts;
typedef vector<vector<long> > vvcounts;

typedef mt19937 base_generator_type;
typedef variate_generator<base_generator_type &, uniform_int<long> > unifrng;
typedef variate_generator<base_generator_type &, normal_distribution<double> > normalrng;

void load_insertion_info(
		istream & is,
		long & start, long & end) {
	string line, ins_str;
	if (getline(is, line) && !is.eof()) {
		istringstream iss(line);
		iss >> ins_str >> start >> end;
	}
}

void load_map_counts(
		istream & is,
		vmers & mers,
		vvcounts & counts) {
	string line;
	int mers_count;
	long mer;
	// load the first line
	getline(is, line);
	istringstream iss(line);
	iss >> mers_count;
	for (int i = 0; i < mers_count; ++i) {
		iss >> mer;
		mers.push_back(mer);
		vcounts counts2;
		counts.push_back(counts2);
	}
	long pos, count;
	while (getline(is, line) && !is.eof()) {
		istringstream iss(line);
		iss >> pos;
		for (int i = 0; i < mers_count; ++i) {
			iss >> count;
			counts[i].push_back(count);
		}
	}
}

long get_count(
		long pos, long size,
		vmers const & mers, vvcounts const & counts) {
	// assumes that the mers are ordered
	vmers::const_iterator mer_itr = lower_bound(mers.begin(), mers.end(), size);
	long index = mer_itr - mers.begin();
	if (mer_itr == mers.end()) {
		return (counts[index - 1][pos] + 1) / 2;	// return 0 iff. the largest mer count is 0
	} else if (*mer_itr == size) {
		return counts[index][pos];
	} else if (mer_itr == mers.begin()) {
		return max(counts[index][pos] * 2 + 1, (long)0);	// always return a positive number
	} else {
		return (counts[index][pos] +
				(long)((double)(mers[index] - size) *
					(double)(counts[index - 1][pos] - counts[index][pos])
					/ (double)(mers[index] - mers[index - 1]) + 0.5));
	}
}

bool set_error_model(
		int const read_type,
		long const & read_length,
		double & error_start,
		double & error_end,
		double & error_inc_rate,
		double & sub_prob,
		double & ins_prob) {
	if (read_type == contig_type::LONG_READ
			|| read_type == contig_type::LONG_READ_PE) {
		error_start = 0.00001;
		error_end = 0.00002;
		error_inc_rate = (error_end - error_start) / (double)read_length;
		sub_prob = 0.8;
		ins_prob = 0.1;
	} else if (read_type == contig_type::MEDIUM_READ
			|| read_type == contig_type::MEDIUM_READ_PE) {
		error_start = 0.003;
		error_end = 0.005;
		error_inc_rate = (error_end - error_start) / (double)read_length;
		sub_prob = 0.17;
		ins_prob = 0.42;
	} else if (read_type == contig_type::SHORT_READ
			|| read_type == contig_type::SHORT_READ_PE) {
		error_start = 0.002;
		error_end = 0.006;
		error_inc_rate = (error_end - error_start) / (double)(read_length);
		sub_prob = 0.333;
		ins_prob = 0.333;
	} else {
		return false;
	}
	return true;
}

void generate_reads(
		int const read_type,
		contig_sets & reads,
		long const & region_size,
		unifrng & pos_rng,
		vmers const & map_mers,
		vvcounts const & map_counts,
		vmers const & map_mers_snp,
		vvcounts const & map_counts_snp,
		double const & total_cost,
		double const & cost_per_bp,
		long const & read_length) {
	double error_start, error_end, error_inc_rate, sub_prob, ins_prob;
	if (!set_error_model(read_type, read_length,
				error_start, error_end, error_inc_rate, sub_prob, ins_prob)) {
		return;
	};
	double num_reads = total_cost / cost_per_bp / (double)(read_length);
	L_(debug) << "Generating " << contig_type::TYPES[read_type] << " reads: "
		<< num_reads << " reads";
	// when num_read < 1, use rnd to decide whether or not to generate a read
	long rnd = pos_rng();
	if (num_reads < 1 && (double)rnd / (double)region_size > num_reads) {
		return;
	}
	// randomly select the start positions of the reads
	vector<long> selected(region_size, 0);
	for (double i = 0; i < num_reads; ++i) {
		long pos = pos_rng();
		selected[pos]++;
	}
	// for each position, generate a set of similar reads
	for (long i = 0; i < region_size; ++i) {
		long same_count = max(get_count(i, read_length,
				map_mers, map_counts), (long)0);
		long similar_count = max(get_count(i, read_length,
				map_mers_snp, map_counts_snp) - same_count, (long)0);
		long sames, similars;
		sames = selected[i];
		similars = 0;
		for (long j = 0; j < same_count - 1; ++j) { // exclude count from the target region
			long pos = pos_rng();
			sames += selected[pos];
		}
		for (long j = 0; j < similar_count; ++j) {
			long pos = pos_rng();
			similars += selected[pos];
		}
		shared_ptr<contig_set> readp(
				new contig_set(read_type, i, read_length));
		// generate read errors for each position
		double error_rate = error_start;
		for (long j = 0; j < read_length; ++j) {
			readp->sames[j] = sames;
			for (long k = 0; k < sames; ++k) {
				double unif_rnd = (double)rand() / ((double)RAND_MAX + (double)1);
				if (unif_rnd <= error_rate) { // read error occurs
					double unif_rnd2 = (double)rand() / ((double)RAND_MAX + (double)1);
					if (unif_rnd2 <= sub_prob) {
						readp->substitutions[j]++;
					} else if (unif_rnd2 <= sub_prob + ins_prob) {
						readp->insertions[j]++;
					} else {
						readp->deletions[j]++;
					}
				}
			}
			readp->similars[j] = similars;
			for (long k = 0; k < similars; ++k) {
				double unif_rnd = (double)rand() / ((double)RAND_MAX + (double)1);
				if (unif_rnd <= error_rate) { // read error
					double unif_rnd2 = (double)rand() / ((double)RAND_MAX + (double)1);
					double unif_rnd3 = (double)rand() / ((double)RAND_MAX + (double)1);
					if (unif_rnd3 > (1.0 / 64.0 * 2.0 / (double)read_length)) {
						if (unif_rnd2 <= sub_prob) {
							readp->substitutions[j]++;
						} else if (unif_rnd2 <= sub_prob + ins_prob) {
							readp->insertions[j]++;
						} else {
							readp->deletions[j]++;
						}
					}
				} else {
					unif_rnd = (double)rand() / ((double)RAND_MAX + (double)1);
					if (unif_rnd <= 2.0 / (double)read_length) { // mismatch
						double unif_rnd2 = (double)rand() / ((double)RAND_MAX + (double)1);
						if (unif_rnd2 <= 0.333333) {
							readp->substitutions[j]++;
						} else if (unif_rnd2 <= 0.666666) {
							readp->insertions[j]++;
						} else {
							readp->deletions[j]++;
						}
					}
				}
			}
			error_rate += error_inc_rate;
		}

		// add to reads vector
		reads.push_back(readp);
	}
}

void generate_pe_reads(
		int const read_type,
		contig_sets & reads,
		long const & region_size,
		unifrng & pos_rng,
		unifrng & end_size_rng,
		normalrng & gap_size_rng,
		vmers const & map_mers,
		vvcounts const & map_counts,
		vmers const & map_mers_snp,
		vvcounts const & map_counts_snp,
		vmers const & map_mers_ref,
		vvcounts const & map_counts_ref,
		vmers const & map_mers_snp_ref,
		vvcounts const & map_counts_snp_ref,
		double const & total_cost,
		double const & cost_per_bp,
		long const & read_length,
		long const & gap_mean,
		long const & gap_sd) {
	double error_start, error_end, error_inc_rate, sub_prob, ins_prob;
	if (!set_error_model(read_type, read_length,
				error_start, error_end, error_inc_rate, sub_prob, ins_prob)) {
		return;
	};
	// The factor of 2 is (approx.) because of the two choices that the
	// other end can come from (up/down-stream of the end that fall into
	// the insertion). Another intepretation is that we want to maintain
	// the ``correct" cost of the read-ends that fall into the insertion
	double num_reads = 2.0 * total_cost / cost_per_bp / (double)(read_length);
	L_(debug) << "Generating " << contig_type::TYPES[read_type] << " reads: "
		<< num_reads << " read-ends";
	// randomly select the start positions of the reads
	vector<long> selected(region_size, 0);
	for (double i = 0; i < num_reads; ++i) {
		long pos = pos_rng();
		selected[pos]++;
	}
	// for each position, generate a set of read ends
	for (long i = 0; i < region_size; ++i) {
		long max_length = 1;
		shared_ptr<contig_set> readp(
				new contig_set(read_type, i, max_length));
		bool read_unique = false;
		for (long k = 0; k < selected[i]; ++k) {
			long end_size = 0;
			switch (read_type) {
				case contig_type::SHORT_READ_PE:
					end_size = read_length / 2;
					break;
				default:
					end_size = end_size_rng();
					break;
			}
			long other_end_size = region_size - other_end_size;
			if (end_size > max_length) {	// resize contig if necessary
				readp->resize(end_size);
				max_length = end_size;
			}
			double error_rate = error_start;
			// decide which end of the paired-end read this belongs to
			double read_end_rng = (double)rand() / ((double)RAND_MAX + (double)1); 
			bool is_left_end = true;
			if (read_end_rng <= 0.5) {	// the left-most end
			} else {	// the right-most end
				error_rate = error_start +
					error_inc_rate * (double)(read_length - end_size);
				is_left_end = false;
			}
			// identify the uniqueness of the paired-end in the reference
			long gap_size = (long)(gap_size_rng());
			long other_end_start;
			if (is_left_end) {
				other_end_start = i + end_size + gap_size;
			} else {
				other_end_start = i - gap_size - other_end_size;
			}
			long other_end_same_count;
			if (other_end_start >= 0 && other_end_start < region_size) {
				other_end_same_count = max(get_count(
							other_end_start, other_end_size,
							map_mers_ref, map_counts_ref), (long)0);
			} else {
				other_end_start = other_end_start % region_size;
				// use the target mapability map to simulate the mapability
				// map of the whole reference genome
				other_end_same_count = max(get_count(
							other_end_start, other_end_size,
							map_mers, map_counts), (long)1);
			}
			
			if (other_end_size >= constants::min_possible_unique_kmer_size
					&& other_end_same_count == 1) {
				// generate read errors for each position
				read_unique = true;
			}

			for (long j = 0; j < end_size; ++j) {
				readp->sames[j]++;
				double unif_rnd = (double)rand() / ((double)RAND_MAX + (double)1);
				if (unif_rnd <= error_rate) { // read error occurs
					double unif_rnd2 = (double)rand() / ((double)RAND_MAX + (double)1);
					if (unif_rnd2 <= sub_prob) {
						readp->substitutions[j]++;
					} else if (unif_rnd2 <= sub_prob + ins_prob) {
						readp->insertions[j]++;
					} else {
						readp->deletions[j]++;
					}
				}
				error_rate += error_inc_rate;
			}
		}
		// add to reads vector
		if (read_unique) {
			readp->max_pe_size = max_length;
		}
		reads.push_back(readp);
	}
}

void combine_reads(
		contig_sets & contigs,
		contig_sets const & reads) {
	if (contigs.empty()) {
		for (unsigned long i = 0; i < reads.size(); ++i) {
			shared_ptr<contig_set> contig(new contig_set(contig_type::CONTIG, i, 0));
			contigs.push_back(contig);
		}
	}

	for (unsigned long i = 0; i < reads.size(); ++i) {
		shared_ptr<contig_set> read = reads[i];
		shared_ptr<contig_set> contig = contigs[i];
		if (read->length == 0) {
			continue;
		}
		if (read->sames[0] + read->similars[0] > 0) {
			contig->max_pe_size = max(contig->max_pe_size, read->max_pe_size);
			if (read->length > contig->length) {
				// extend the contig
				for (long j = contig->length; j < read->length; ++j) {
					contig->sames.push_back(0);
					contig->similars.push_back(0);
					contig->substitutions.push_back(0);
					contig->insertions.push_back(0);
					contig->deletions.push_back(0);
				}
				contig->length = read->length;
			}
			// modify the contig
			for (long j = 0; j < read->length; ++j) {
				contig->sames[j] += read->sames[j];
				contig->similars[j] += read->similars[j];
				contig->substitutions[j] += read->substitutions[j];
				contig->insertions[j] += read->insertions[j];
				contig->deletions[j] += read->deletions[j];
			}
		}
	}
}

void copy_contig_info(shared_ptr<contig_set> const & src,
		shared_ptr<contig_set> const & dest) {
	dest->start = src->start;
	dest->length = src->length;
	dest->sames = src->sames;
	dest->similars = src->similars;
	dest->substitutions = src->substitutions;
	dest->insertions = src->insertions;
	dest->deletions = src->deletions;
	dest->max_pe_size = src->max_pe_size;
}

void update_contig_info(shared_ptr<contig_set> const & src,
		shared_ptr<contig_set> const & dest,
		long const & start, long const & size) {
	for (long i = 0; i < size; ++i) {
		long j = start + i;
		dest->sames[j] += src->sames[i];
		dest->similars[j] += src->similars[i];
		dest->substitutions[j] += src->substitutions[i];
		dest->insertions[j] += src->insertions[i];
		dest->deletions[j] += src->deletions[i];
	}
	dest->max_pe_size = max(dest->max_pe_size, min(src->max_pe_size, size));
}

void extend_contig(shared_ptr<contig_set> const & dest,
		long const & size) {
	dest->length += size;
	for (long i = 0; i < size; ++i) {
		dest->sames.push_back(0);
		dest->similars.push_back(0);
		dest->substitutions.push_back(0);
		dest->insertions.push_back(0);
		dest->deletions.push_back(0);
	}
}

void build_contigs(
		contig_sets & contigs,
		contig_sets const & reads,
		long const & overlap_threshold,
		vmers const & map_mers,
		vvcounts const & map_counts,
		unifrng & pos_rng) {
	shared_ptr<contig_set> cur_contig(new contig_set(contig_type::CONTIG, 0, 0)); 
	bool cur_contig_initialized = false;
	long pos = 0;
	bool pe_expansion_flag = false;	// becomes true immediately after an expansion has been taken place because of a paired-end read
	while (pos < (long)reads.size()) {
		bool local_pe_expansion_flag = pe_expansion_flag;
		pe_expansion_flag = false;
		shared_ptr<contig_set> read = reads[pos];
		long cur_end = cur_contig->start + cur_contig->length;
		if (read->length > 0) {
			long read_start = read->start;
			long read_end = pos + read->length;
			if (!cur_contig_initialized) {
				// copy read information
				copy_contig_info(read, cur_contig);
				cur_contig_initialized = true;
			} else if (cur_end - read_start < overlap_threshold) {
				// save current contig and create a new one
				contigs.push_back(cur_contig);
				shared_ptr<contig_set> tmp_contig(new contig_set(contig_type::CONTIG, read_start, read->length)); 
				cur_contig = tmp_contig;
				copy_contig_info(read, cur_contig);
			} else if (read_end <= cur_end) {
				// add read info into the contig
				update_contig_info(read, cur_contig,
						read_start - cur_contig->start, read_end - read_start);
			} else {
				// try to extend the contig
				long pe_length = read->max_pe_size;
				if (pe_length > cur_end - read_start && !local_pe_expansion_flag) {
					// always extend for localized paired-end reads
					extend_contig(cur_contig, read_start + pe_length - cur_end);
					update_contig_info(read, cur_contig,
							read_start - cur_contig->start, pe_length);
					// set flag
					pe_expansion_flag = true;
					// try to extend with the same read again
					continue;
				}
				long count_overlap_target = get_count(
						read->start, cur_end - read_start,
						map_mers, map_counts);
				long count_read_target = get_count(
						read->start, read->length,
						map_mers, map_counts);
				if (count_read_target >= count_overlap_target) {
					if (read->length > 30 || read->sames[0] + read->similars[0] > 2) { 
						// extend
						extend_contig(cur_contig, read_end - cur_end);
						if (local_pe_expansion_flag) {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start + pe_length,
									read_end - read_start - pe_length);
						} else {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start,
									read_end - read_start);
						}
					}
				} else {
					long num_diff_loci = count_overlap_target - count_read_target;
					long num_misleading_reads = 0;
					long consistent_extension = read->length - (cur_end - read_start);
					while (count_read_target < count_overlap_target &&
							consistent_extension > 0) {
						consistent_extension--;
						count_read_target = get_count(
								read->start, cur_end - read_start + consistent_extension,
								map_mers, map_counts);
					}
					for (long j = 0; j < num_diff_loci; ++j) {
						long tmp_pos = pos_rng();
						if (reads[tmp_pos]->length >= read->length) {
							num_misleading_reads++;
						}
					}
					if (num_misleading_reads * 5 < read->sames[0] + read->similars[0]) {
						// a lot more correct reads, extend
						extend_contig(cur_contig, read_end - cur_end);
						if (local_pe_expansion_flag) {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start + pe_length,
									read_end - read_start - pe_length);
						} else {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start,
									read_end - read_start);
						}
					} else if (num_misleading_reads / 20 > read->sames[0] + read->similars[0]) {
						// too many misleading reads, save current contig and create a new one
						contigs.push_back(cur_contig);
						shared_ptr<contig_set> tmp_contig(new contig_set(contig_type::CONTIG, read_start, read->length)); 
						cur_contig = tmp_contig;
						copy_contig_info(read, cur_contig);
					} else {
						// intermediate case, extend the consistent part
						extend_contig(cur_contig, consistent_extension);
						if (local_pe_expansion_flag) {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start + pe_length,
									cur_end - read_start + consistent_extension - pe_length);
						} else {
							update_contig_info(read, cur_contig,
									read_start - cur_contig->start,
									cur_end - read_start + consistent_extension);
						}
					}
				}
			}
		} else {
			// do nothing
		}
		++pos;
	}
	contigs.push_back(cur_contig);
}

bool identify_insertion(
		contig_sets const & contigs,
		long const & insertion_start,
		long const & insertion_end,
		vmers const & map_mers,
		vvcounts const & map_counts) {
	bool insertion_detected = false;
	for (unsigned long i = 0; i < contigs.size(); ++i) {
		shared_ptr<contig_set> contig = contigs[i];
		if (contig->start < insertion_start &&
				contig->start + contig->length > insertion_start) {
			long count_left_ref = get_count(
					contig->start, insertion_start - contig->start,
					map_mers, map_counts);
			long count_read_ref = get_count(
					contig->start, contig->length,
					map_mers, map_counts);
			if (count_left_ref <= 1 && count_read_ref == 0) {
				insertion_detected = true;
				break;
			}
		}
		if (contig->start < insertion_end &&
				contig->start + contig->length > insertion_end) {
			long count_right_ref = get_count(
					insertion_end, contig->start + contig->length - insertion_end,
					map_mers, map_counts);
			long count_read_ref = get_count(
					contig->start, contig->length,
					map_mers, map_counts);
			if (count_right_ref <= 1 && count_read_ref == 0) {
				insertion_detected = true;
				break;
			}
		}
	}
	return insertion_detected;
}

long count_detected(shared_ptr<contig_set> const & contig,
		long const & start,
		long const & end) {
	long detected = 0;
	for (long i = start; i < end; ++i) {
		long sames = contig->sames[i];
		long similars = contig->similars[i];
		long substitutions = contig->substitutions[i];
		long insertions = contig->insertions[i];
		long deletions = contig->deletions[i];
		if ((contig->length > 30 ||
					sames + similars > 2) &&
				sames + similars > 3 * (substitutions + insertions + deletions)) {
			detected++;
		}
	}
	return detected;
}

void count_detection(
		contig_sets const & contigs,
		long const & insertion_start,
		long const & insertion_end,
		long & detected) {
	L_(debug) << "Counting total detection...";
	long cur_end = insertion_start;
	for (unsigned long i = 0; i < contigs.size(); ++i) {
		shared_ptr<contig_set> contig = contigs[i];

		long local_detection = 0;
		if (contig->start <= insertion_start &&
				contig->start + contig->length >= insertion_start) {
			local_detection = count_detected(contig,
					insertion_start - contig->start,
					min(contig->length, insertion_end - contig->start));
			detected += local_detection;
			cur_end = min(contig->start + contig->length, insertion_end);
		} else if (contig->start <= insertion_end &&
				contig->start + contig->length >= insertion_end) {
			local_detection = count_detected(contig,
					max(cur_end - contig->start, (long)0),
					insertion_end - contig->start);
			detected += local_detection;
			cur_end = insertion_end;
		} else {
			// For those paired-end-read-localized contigs inside the insertion,
			// we can still identify them even if the breakpoints of the novel
			// insertion is not detected
			if (contig->max_pe_size > constants::min_pe_based_contig_max_pe_length) {
				local_detection = count_detected(contig,
						max(cur_end - contig->start, (long)0),
						min(contig->length, insertion_end - contig->start));
				detected += local_detection;
				cur_end = max(cur_end, contig->start + min(contig->length,
							insertion_end - contig->start));
			}
		}
		if (local_detection > 0) {
			L_(debug) << "Contig [localized]: [" << contig->start << ", "
				<< (contig->start + contig->length) << ")";
			L_(debug) << "Additional detection: " << local_detection << " bp";
			L_(debug) << "Current detection end: " << cur_end;
		} else {
			L_(debug2) << "Contig [not localized]: [" << contig->start << ", "
				<< (contig->start + contig->length) << ")";
		}
	}
}

void print_contig_sets(contig_sets const & contig_sets) {
	BOOST_FOREACH(shared_ptr<contig_set> const & read, contig_sets) {
		if (read->length == 0 || read->sames[0] + read->similars[0] == 0) {
			continue;
		}
		L_(debug3) << contig_type::TYPES[read->type] << ": start: "
			<< read->start << ", length: " << read->length
			<< ", max_pe_size: " << read->max_pe_size
			<< ", quantity: " << read->sames[0] << " sames, "
			<< read->similars[0] << " similars"
			<< ", subst: " << accumulate(read->substitutions.begin(),
					read->substitutions.end(), 0)
			<< ", ins: " << accumulate(read->insertions.begin(),
					read->insertions.end(), 0)
			<< ", del: " << accumulate(read->deletions.begin(),
					read->deletions.end(), 0);
	}
}

#endif
